import logging
from ..base_config import BaseClassifierConfig

logger = logging.getLogger(__name__)


class ImageTextClassifierConfig(BaseClassifierConfig):
    """
    ImageText Classifier Config.

    Args:
        text_prompt (str): Text prompt for text classification, default: "{}의 상품 종류는?".
        image_backbone_name (str): Image backbone name, default: "eva_large_patch14_196.in22k_ft_in22k_in1k".
        text_backbone_name (str): Text backboname, default: "EleutherAI/polyglot-ko-1.3b".
        classifier_learning_reate (float): Classifier learning rate, default: 1e-3.
        image_backbone_learning_rate (float): Image backbone learning rate, default: 4e-5.
        text_backbone_learning_rate (float): Text backbone learning rate, default: 3e-4.
        image_freeze (bool): Freeze image backbone, default: False.
        text_freeze (bool): Freeze text backbone, default: False.
        image_lora (bool): Use LoRA for image backbone, default: True.
        text_lora (bool): Use LoRA for text backbone, default: True.
        image_lora_r (int): LoRA r for image backbone, default: 16.
        text_lora_r (int): LoRA r for text backbone, default: 16.
        image_lora_alpha (int): LoRA alpha for image backbone, default: 16.
        text_lora_alpha (int): LoRA alpha for text backbone, default: 16.
        image_lora_target_modules (list): LoRA target modules for image backbone, default: ["qkv"].
        text_lora_target_modules (list): LoRA target modules for text backbone, default: ["query_key_value"].
        image_lora_dropout (float): LoRA dropout for image backbone, default: 0.1.
        text_lora_dropout (float): LoRA dropout for text backbone, default: 0.1.
        auto_augment (str): Auto augment for image backbone, default: "rand-m9-mstd0.5-inc1".
        crop_ratio (float): Minimum crop ratio for image backbone when augmentation image, default: 0.9.
    """

    def __init__(
        self,
        text_prompt: str = "{}의 상품 종류는?",
        image_backbone_name: str = "eva_large_patch14_196.in22k_ft_in22k_in1k",
        text_backbone_name: str = "EleutherAI/polyglot-ko-1.3b",
        classifier_learning_rate: float = 1e-3,
        image_backbone_learning_rate: float = 4e-5,
        text_backbone_learning_rate: float = 3e-4,
        image_freeze: bool = False,
        text_freeze: bool = False,
        image_lora: bool = True,
        text_lora: bool = True,
        image_lora_r: int = 16,
        text_lora_r: int = 16,
        image_lora_alpha: int = 16,
        text_lora_alpha: int = 16,
        image_lora_target_modules: list = None,
        text_lora_target_modules: list = None,
        image_lora_dropout: float = 0.1,
        text_lora_dropout: float = 0.1,
        auto_augment: str = "rand-n3-mstd1",
        crop_ratio: float = 0.9,
        **kwargs
    ):
        super().__init__(**kwargs)

        self.text_prompt = text_prompt
        self.image_backbone_name = image_backbone_name
        self.text_backbone_name = text_backbone_name
        self.classifier_learning_rate = classifier_learning_rate
        self.image_backbone_learning_rate = image_backbone_learning_rate
        self.text_backbone_learning_rate = text_backbone_learning_rate

        # Training Config
        self.image_freeze = image_freeze
        self.text_freeze = text_freeze
        self.image_lora = image_lora
        self.text_lora = text_lora
        if self.image_freeze and self.image_lora:
            raise ValueError(
                "freezing backbone and using lora for image backbone cannot be True at the same time"
            )
        if self.text_freeze and self.text_lora:
            raise ValueError(
                "freezing backbone and using lora for text backbone cannot be True at the same time"
            )

        self.image_lora_r = image_lora_r
        self.text_lora_r = text_lora_r
        self.image_lora_alpha = image_lora_alpha
        self.text_lora_alpha = text_lora_alpha
        if image_lora_target_modules is None:
            self.image_lora_target_modules = ["qkv"]
        else:
            self.image_lora_target_modules = image_lora_target_modules
        if text_lora_target_modules is None:
            self.text_lora_target_modules = ["query_key_value"]
        else:
            self.text_lora_target_modules = text_lora_target_modules
        self.image_lora_dropout = image_lora_dropout
        self.text_lora_dropout = text_lora_dropout
        # Image Augmentation
        self.auto_augment = auto_augment
        self.crop_ratio = crop_ratio

        # Boolean to check if the model is loaded from transformers or timm, will be set in trainer
        self.text_timm = False
        self.image_timm = None



import os

os.environ["TOKENIZERS_PARALLELISM"] = "false"
import copy
import torch

torch.backends.cudnn.benchmark = False
import logging
from typing import List, Union

import pandas as pd
import pytorch_lightning as pl
from pytorch_lightning.callbacks import ModelCheckpoint

from .trainer_util import set_seed, sample_stratified

logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)


def check_dataset(
    dataset: Union[pd.DataFrame, List[List]], column_list: List[str], train_config=None
):
    """
    Check if the dataset is in the correct format
    Args:
        dataset (pd.DataFrame or list): dataset
        column_list (list): list of column names (must be in order of text, label)
    Returns:
        output dataset
    """

    if train_config != None:
        if train_config.problem_type == "multi_label_classification":
            try:
                dataset["label"] = dataset["label"].apply(lambda x: eval(x))
            except:
                raise ValueError(f"cannot convert dataset")

    if type(dataset) == pd.DataFrame:
        if len(dataset.columns) != len(column_list):
            raise ValueError(
                f"For DataFrame input, dataset must have {len(column_list)} columns. Got {len(dataset.columns)} columns"
            )
        for idx, col in enumerate(column_list):
            if dataset.columns[idx] != col:
                logger.warning(
                    f"Column {idx} is {dataset.columns[idx]}, but expected {col}"
                )
            dataset.rename(columns={dataset.columns[idx]: col}, inplace=True)
    elif type(dataset) == list:
        if len(dataset[0]) != len(column_list):
            raise ValueError(
                f"For list input, dataset must have {len(column_list)} columns. Got {len(dataset[0])} columns"
            )
        else:
            logger.warning(
                f"For list input, column names will be set to default: {column_list}"
            )
    else:
        raise ValueError(
            f"Dataset must be either pd.DataFrame or list. Got {type(dataset)}"
        )

    if dataset["label"].dtype == "object":
        if not type(dataset["label"][0]) == list:
            raise ValueError(
                f"Multi Label column must be list type. Got {type(dataset['label'][0])}"
            )
        elif not type(dataset["label"][0][0]) == int:
            raise ValueError(
                f"Element of Multi Label column must be int type. Got {type(dataset['label'][0][0])}"
            )

    elif dataset["label"].dtype != "int64":
        raise ValueError(
            f"Label column must be int64 dtype. Got {dataset['label'].dtype}"
        )

    return dataset


class BaseClassifier:
    """
    Base class for all classifiers, including text, image, and image-text classifiers
    Args:
        train_dataset (pd.DataFrame or list): training dataset
        val_dataset (pd.DataFrame or list): validation dataset
        feedback_dataset (pd.DataFrame or list): feedback dataset (optional)
        test_dataset (pd.DataFrame or list): test dataset (optional)
        label_name_list (list): list of label names (optional)
        train_config (ClassifierConfig): training config (optional)
    """

    def __init__(
        self,
        train_dataset: Union[pd.DataFrame, List[List]],
        val_dataset: Union[pd.DataFrame, List[List]],
        feedback_dataset: Union[pd.DataFrame, List[List]] = None,
        test_dataset: Union[pd.DataFrame, List[List]] = None,
        label_name_list: List = None,
        train_config=None,
    ):
        from .text_classifier import TextClassifierConfig
        from .image_classifier import ImageClassifierConfig
        from .image_text_classifier import ImageTextClassifierConfig

        self.train_config = train_config
        # Set random seed
        set_seed(self.train_config.seed)

        train_dataset = copy.deepcopy(train_dataset)
        val_dataset = copy.deepcopy(val_dataset)
        feedback_dataset = copy.deepcopy(feedback_dataset)
        test_dataset = copy.deepcopy(test_dataset)
        if type(train_config) == TextClassifierConfig:
            train_dataset = check_dataset(
                train_dataset, ["text", "label"], train_config
            )
            val_dataset = check_dataset(val_dataset, ["text", "label"], train_config)
            if feedback_dataset is not None:
                feedback_dataset = check_dataset(
                    feedback_dataset, ["text", "label"], train_config
                )
            if test_dataset is not None:
                test_dataset = check_dataset(
                    test_dataset, ["text", "label"], train_config
                )
            else:
                logger.warning(
                    "test_dataset is None. Using val_dataset as test_dataset"
                )
                test_dataset = copy.deepcopy(val_dataset)

        elif type(train_config) == ImageClassifierConfig:
            train_dataset = check_dataset(
                train_dataset, ["img_path", "label"], train_config
            )
            val_dataset = check_dataset(
                val_dataset, ["img_path", "label"], train_config
            )
            if feedback_dataset is not None:
                feedback_dataset = check_dataset(
                    feedback_dataset, ["img_path", "label"], train_config
                )
            if test_dataset is not None:
                test_dataset = check_dataset(
                    test_dataset, ["img_path", "label"], train_config
                )
            else:
                logger.warning(
                    "test_dataset is None. Using val_dataset as test_dataset"
                )
                test_dataset = copy.deepcopy(val_dataset)

        elif type(train_config) == ImageTextClassifierConfig:
            train_dataset = check_dataset(
                train_dataset, ["img_path", "text", "label"], train_config
            )
            val_dataset = check_dataset(
                val_dataset, ["img_path", "text", "label"], train_config
            )
            if feedback_dataset is not None:
                feedback_dataset = check_dataset(
                    feedback_dataset, ["img_path", "text", "label"], train_config
                )
            if test_dataset is not None:
                test_dataset = check_dataset(
                    test_dataset, ["img_path", "text", "label"], train_config
                )
            else:
                logger.warning(
                    "test_dataset is None. Using val_dataset as test_dataset"
                )
                test_dataset = copy.deepcopy(val_dataset)
        else:
            raise ValueError(f"Not supported train_config type: {type(train_config)}")

        ## If feedback_dataset exists, concat feedback dataset and stratified sample from train dataset
        if feedback_dataset is not None:
            ##### 요부분 수정 필요! @김태윤 ############
            sampled_train_dataset = sample_stratified(
                train_dataset,
                # self.train_config.feedback_iteration * len(feedback_dataset),
                self.train_config.feedback_train_sample_per_label,
                self.train_config.problem_type,
            )
            # Add feedback label for distinguishing feedback data
            sampled_train_dataset["feedback_flag"] = [0] * len(sampled_train_dataset)
            repeated_feedback_dataset = pd.concat(
                [feedback_dataset] * self.train_config.feedback_iteration,
                ignore_index=True,
            )
            repeated_feedback_dataset["feedback_flag"] = [1] * len(
                repeated_feedback_dataset
            )
            self.train_dataset = pd.concat(
                [sampled_train_dataset, repeated_feedback_dataset], ignore_index=True
            )
        else:
            self.train_dataset = train_dataset
        self.val_dataset = val_dataset
        self.test_dataset = test_dataset

        # Check number of class when training config num class value is default value.
        if self.train_config.num_class == 0:
            if label_name_list is not None:
                logger.info("Get number of class with the length of label name list.")
                self.train_config.num_class = len(label_name_list)
                self.train_config.label_name_list = label_name_list
            else:
                label_list = (
                    list(map(lambda x: int(x), train_dataset["label"]))
                    + list(map(lambda x: int(x), val_dataset["label"]))
                    + list(map(lambda x: int(x), test_dataset["label"]))
                )
                self.train_config.num_class = max(label_list) + 1
                logger.warning(
                    f"Since training config does not get the number of class, trainer estimate the number of class with input dataset, estimated number of class is {self.train_config.num_class} "
                )

        else:
            if label_name_list is not None:
                if self.train_config.num_class != len(label_name_list):
                    raise ValueError(
                        f"Number of class in training config {self.train_config.num_class} does not match with length of label_name_list {len(label_name_list)}."
                    )
                self.train_config.label_name_list = label_name_list

        self.trainer = None

    def base_train(self):
        """
        Base training function
        """
        if self.train_config.problem_type == "single_label_classification":
            if self.train_config.monitor_topk == "total":
                monitor_acc = "val/tot_acc"
                if self.train_config.val_check_interval:
                    filename = (
                        "epoch{epoch:02d}-step{step}-val_totacc={val/tot_acc:.4f}"
                    )
                else:
                    filename = "epoch{epoch:02d}-val_totacc={val/tot_acc:.4f}"
                checkpoint_callback = ModelCheckpoint(
                    monitor=monitor_acc,
                    mode="max",
                    save_top_k=self.train_config.save_ckpt_topk,
                    filename=filename,
                    auto_insert_metric_name=False,
                )
            else:
                monitor_acc = f"val/acc{self.train_config.monitor_topk}"
                if self.train_config.val_check_interval:
                    filename = (
                        "epoch{epoch:02d}-step{step}-val_acc={val/monitor_acc:.4f}"
                    )
                else:
                    filename = "epoch{epoch:02d}-val_acc={val/monitor_acc:.4f}"
                checkpoint_callback = ModelCheckpoint(
                    monitor=monitor_acc,
                    mode="max",
                    save_top_k=self.train_config.save_ckpt_topk,
                    filename=filename,
                    auto_insert_metric_name=False,
                )
        else:
            monitor_f1 = f"val/f1"
            if self.train_config.val_check_interval:
                filename = "epoch{epoch:02d}-step{step}-val_f1={val/f1:.4f}"
            else:
                filename = "epoch{epoch:02d}-val_f1={val/f1:.4f}"
            checkpoint_callback = ModelCheckpoint(
                monitor=monitor_f1,
                mode="max",
                save_top_k=self.train_config.save_ckpt_topk,
                filename=filename,
                auto_insert_metric_name=False,
            )

        # Define trainer
        self.trainer = pl.Trainer(
            accelerator=self.train_config.device,
            devices=self.train_config.use_device_list,
            precision=self.train_config.precision,
            max_epochs=self.train_config.max_epochs,
            callbacks=[checkpoint_callback],
            num_sanity_val_steps=self.train_config.num_sanity_val_steps,
            check_val_every_n_epoch=self.train_config.check_val_every_n_epoch,
            val_check_interval=self.train_config.val_check_interval,
            default_root_dir=self.train_config.output_dir,
            strategy="ddp_find_unused_parameters_true",
        )

        self.trainer.fit(self.train_module, self.data_module)

        print(
            "Best ckpt saved path: ", self.trainer.checkpoint_callback.best_model_path
        )

    # Test the best trained model with test dataset
    def test(self):
        """
        Test the best trained model with test dataset, if test dataset is None, test with val dataset.
        """
        if self.trainer is None:
            raise ValueError(
                "You should train the trainer before inference. E.g. TextClassifier().train()"
            )
        self.trainer.test(ckpt_path="best", dataloaders=self.data_module)



import logging
import torch.nn as nn
from typing import Union, List

from ..losses import LOSSES

logger = logging.getLogger(__name__)


class BaseClassifierConfig:
    """
    Base Classifier Config Class

    Args:
        batch_size (int): batch size per GPU, default: 32
        max_epochs (int): max epoch for training, default: 10
        feedback_iteration (int): feedback iteration, only used for feedback_training, default: 1.
        feedback_train_sample_per_label (int): feedback train sample number per each label, only used for feedback_training, default: 400.
        e.g. 100 feed back iterations and 5 label, 10 feedback sample -> 100 * 5 * 10 = 5000 samples will be used in train dataset.
        seed (int): random seed, default: 42
        num_class (int): number of class, default: 0
        output_dir (str): output directory, default: 'tmp_classifier'
        device (str): device name, e.g. 'cpu', 'gpu', default: 'gpu'
        use_device_list (Union[str, List[int]]): device list, e.g. 'auto', [0, 1], default: 'auto'
        num_workers (int): number of workers, default: 4
        fc_classifier (Union[str, nn.Module]): fully connected classifier, e.g. 'Base2LinearClassifier', default: 'Base2LinearClassifier'
        label_smoothing_factor (float): label smoothing factor, default: 0.1
        feedback_ema_decay (float): feedback ema decay, default: 0.9
        gradient_accumulation_steps (int): gradient accumulation steps, default: 1
        print_topk_list (tuple): topk tuple for printing, default: (1, 3, 5)
        monitor_topk (Union[int, str]): topk for monitoring, default: 1
        save_ckpt_topk (int): topk for saving checkpoint, default: 3
        check_val_every_n_epoch (int): check validation every n epoch, default: 1
        val_check_interval(int): check validation every n train steps, also saving model checkpoint. default: None,
        num_sanity_val_steps (int): number of sanity validation steps, default: 2
        precision (str): precision, e.g. '16-mixed', '32', default: '16-mixed'
        label_name_list (list): label name list, default: None
        criterion (str): criterion name, e.g. 'cross_entropy', 'focal_loss', default: 'cross_entropy'
        problem_type (str): problem type, e.g. 'single_label_classification', 'multi_label_classification', default: 'single_label_classification'
        lr_scheduler (str): lr scheduler, e.g. 'linear_warmup', 'constant', default: 'linear_warmup'
    """

    def __init__(
        self,
        batch_size: int = 32,
        max_epochs: int = 10,
        feedback_iteration: int = 1,
        feedback_train_sample_per_label: int = 400,
        seed: int = 42,
        num_class: int = 0,
        output_dir: str = "tmp_classifier",
        device: str = "gpu",
        use_device_list: Union[str, List[int]] = "auto",
        num_workers: int = 4,
        fc_classifier: Union[str, nn.Module] = "BaseLinearClassifier",
        label_smoothing_factor: float = 0.1,
        feedback_ema_decay: float = 0.9,
        gradient_accumulation_steps: int = 1,
        print_topk_list: tuple = (1, 3, 5),
        monitor_topk: Union[int, str] = 1,
        save_ckpt_topk: int = 3,
        check_val_every_n_epoch: int = 1,
        val_check_interval: int = None,
        num_sanity_val_steps: int = 2,
        precision: str = "16-mixed",
        label_name_list: list = None,
        criterion: str = "cross_entropy",
        problem_type: str = "single_label_classification",
        lr_scheduler: str = "linear_warmup",
    ):
        # Batch size per GPU
        self.batch_size = batch_size
        # Max epoch for training
        self.max_epochs = max_epochs
        # Feedback iteration, only used for feedback_training
        self.feedback_iteration = feedback_iteration
        self.feedback_train_sample_per_label = feedback_train_sample_per_label
        self.seed = seed
        # Num class will be define and be checked at trainer class
        self.num_class = num_class
        self.output_dir = output_dir
        self.device = device
        # Can be GPU Device number e.g. [0, 1]
        self.use_device_list = use_device_list
        self.num_workers = num_workers
        # Can be classifier name string or defined nn.Moudle
        self.fc_classifier = fc_classifier
        self.label_smoothing_factor = label_smoothing_factor
        self.feedback_ema_decay = feedback_ema_decay
        self.gradient_accumulation_steps = gradient_accumulation_steps

        self.print_topk_list = list(print_topk_list)
        if monitor_topk != "total" and monitor_topk not in print_topk_list:
            raise ValueError("monitor_topk must be in print_topk_list or 'total'")

        self.monitor_topk = monitor_topk
        self.save_ckpt_topk = save_ckpt_topk
        self.check_val_every_n_epoch = check_val_every_n_epoch
        self.val_check_interval = val_check_interval
        self.num_sanity_val_steps = num_sanity_val_steps
        self.precision = precision
        self.label_name_list = label_name_list

        if problem_type not in [
            "single_label_classification",
            "multi_label_classification",
        ]:
            raise ValueError(
                "problem_type must be 'single_label_classification' or 'multi_label_classification'"
            )
        self.problem_type = problem_type

        if criterion not in LOSSES.keys():
            raise ValueError(f"criterion must be in {LOSSES.keys()}")

        if problem_type == "multi_label_classification":
            if criterion not in ["binary_cross_entropy", "binary_focal_loss"]:
                raise ValueError(
                    f"criterion must be in ['binary_cross_entropy', 'binary_focal_loss'] for multi_label_classification"
                )

        if criterion == "focal_loss" and label_smoothing_factor > 0:
            criterion = "focal_loss_with_label_smoothing"
        self.criterion = criterion

        if lr_scheduler not in ["linear_warmup", "constant"]:
            raise ValueError(f"lr_scheduler must be in ['linear_warmup', 'constant']")
        self.lr_scheduler = lr_scheduler

    def to_dict(self):
        """
        Convert class attributes to dictionary
        """
        return self.__dict__
